from .base_chunker import BaseChunker
from typing import List
import textwrap

class TextChunker(BaseChunker):
    """ Primary chunker for handling most textual data """

    def __init__(self, max_chars: int = 500):
        self.max_chars = max_chars

    def chunk(self, data: str) -> List[str]:
        # Use textwrap to break the text into chunks without breaking words
        chunks = textwrap.wrap(data, width=self.max_chars, break_long_words=False)
        #logger.info(f"Chunked data into {len(chunks)} chunks using TextChunker.")
        return chunks



# base_bot.py

class BaseBot:
    def __init__(self, embedder):
        self.embedder = embedder

    def add_data(self, data_type, data_content):
        """Add data to the system."""
        # Use the embedder to process and add the data
        self.embedder.add(data_type, data_content)

    def query(self, question):
        """Query the system with a specific question."""
        # Use the embedder to process the query and get a response
        response = self.embedder.query(question)
        return response

    def start(self):
        """Start the bot's functionality."""
        raise NotImplementedError("This method should be implemented by subclasses.")



from abc import ABC, abstractmethod
from typing import List, Tuple, Optional, Union
import hashlib

class VectorDatabase(ABC):

    @abstractmethod
    def store_vectors(self, vectors: List[List[float]], metadata_list: List[dict], chunks: List[Union[str, bytes]], store:bool) -> None:
        """
        Store vectors along with associated metadata.

        :param vectors: List of embeddings (dense vectors) to store.
        :param metadata: List of metadata associated with each vector.
        """
        pass

    @abstractmethod
    def search_vectors(self, query_vector: List[float], top_k: int = 5) -> List[Tuple[int, float, dict]]:
        """
        Perform a similarity search and retrieve the top_k most similar vectors to the query_vector.

        :param query_vector: The vector for which similar vectors need to be found.
        :param top_k: Number of top similar vectors to return.
        :return: List of tuples with (index, similarity_score, metadata).
        """
        pass

    @abstractmethod
    def update_vector(self, index: int, new_vector: List[float], new_metadata: Optional[dict] = None) -> None:
        """
        Update a vector in the database.

        :param index: The index of the vector to be updated.
        :param new_vector: The new vector.
        :param new_metadata: The new metadata. If None, the metadata remains unchanged.
        """
        pass

    @abstractmethod
    def delete_vector(self, index: int) -> None:
        """
        Remove a vector from the database.

        :param index: The index of the vector to be removed.
        """
        pass

    @staticmethod
    def content_hash(input: Union[str, bytes]) -> bytes:
        "Hash content for deduplication. Override to change hashing behavior."
        if isinstance(input, str):
            input = input.encode("utf8")
        return hashlib.sha256(input).digest()



from nchain.pipeline import DataPipeline
from nchain.loaders import ArxivLoader, PdfLoader, SQLiteLoader
from nchain.chunkers import TextChunker
from nchain.vectordb import SQLiteVectorDB
from nchain.utils.detection import detect_data_type
from nchain.utils.sqlite_logger import logger
from nchain.embedders import SentenceTransformersEmbedder
from nchain import user_dir
from time import time
import sqlite_utils

class App:
    def __init__(self, db_path=":memory:", vectordb_path: str = str(user_dir() / "embeddings.db"), indexdb_path: str = str(user_dir() / "index.ann")):
        self.db_path = db_path
        self.db = sqlite_utils.Database(db_path)
        self.db["query_results"].create({
            "query": str,
            "data_type": str,
            "embedding_id": int,
            "score": float,
            "content": str,
            "metadata": str,
            "created": int
        }, pk=("query", "data_type", "embedding_id"), if_not_exists=True)
        self.vectordb_path = vectordb_path
        self.embedder = SentenceTransformersEmbedder()
        # Loaders
        self.loaders = {
            "ARXIV_PAPER": ArxivLoader(self.db_path),
            "PDF_FILE": PdfLoader(),
            "SQLITE_DATA": SQLiteLoader(self.db_path)
        }

        self.vectordbs = {
            "ARXIV_PAPER": SQLiteVectorDB(dimension=self.embedder.dimension, indexdb_path=indexdb_path, db_path=vectordb_path, collection="axiv_paper")
        }
        self.chunkers = {
            "ARXIV_PAPER": TextChunker(max_chars=1000)
        }

    def add(self, source: str, data_type: str = None):
        """Adds a new source to the database."""
        if data_type is None:
            data_type = detect_data_type(source)

        loader = self.loaders.get(data_type)
        if not loader:
            raise ValueError(f"Unsupported data type: {data_type}", f"Supported data types: {self.loaders.keys()}")
        chunker = self.chunkers.get(data_type)
        if not chunker:
            raise ValueError(f"Unsupported data type: {data_type}", f"Supported data types: {self.chunkers.keys()}")
        vectordb = self.vectordbs.get(data_type)
        if not vectordb:
            raise ValueError(f"Unsupported data type: {data_type}", f"Supported data types: {self.vectordbs.keys()}")

        # Create a DataPipeline instance
        pipeline = DataPipeline(loader=loader, embedder=self.embedder, chunker=chunker, vectordb=vectordb)
        # Process the source using the pipeline
        pipeline.process(source)
        logger.info(f"Added source: {source} to the database.")

    def query(self, user_query: str, top_k: int = 5):
        """Query the database for similar embeddings."""
        query_embedding = self.embedder.embed(user_query)  # Use the correct method

        results = []
        for data_type, vectordb in self.vectordbs.items():
            #Search for similar embeddings in the SQLiteVectorDB
            similar_entries = vectordb.search_vectors(query_embedding, top_k)

            # Store Query Result
            for entry in similar_entries:
                self.db['query_results'].upsert(
                     {"query": user_query, "data_type": data_type, "embedding_id": entry.id, "score": entry.score, 
                      "content":entry.content, "metadata": entry.metadata,
                      "created": int(time())}, pk=("query", "data_type", "embedding_id"))

            results.extend(similar_entries)
        
        return results



import os
from pathlib import Path

def user_dir():
    chain_user_path = os.environ.get("NANOCHAIN_USER_PATH")
    if chain_user_path:
        path = Path(chain_user_path)
    else:
        path = Path(os.path.expanduser("~/.nchain"))
    path.mkdir(exist_ok=True, parents=True)
    return path



import sqlite_utils
from pydantic import BaseModel, create_model
from typing import Type

class DynamicModelGenerator:
    def __init__(self, db_path: str):
        """
        Initialize the DynamicModelGenerator with a path to the SQLite database.

        :param db_path: Path to the SQLite database file.
        """
        self.db = sqlite_utils.Database(db_path)
    
    # TODO: Figure out how to dynamically create a Pydantic model
    def generate_model(self, table_name: str) -> Type[BaseModel]:
        """
        Dynamically generate a Pydantic model based on the structure (schema) of an SQLite table.

        :param table_name: Name of the SQLite table for which the model needs to be generated.
        :return: Dynamically generated Pydantic model class for the specified table.
        """
        table = self.db[table_name]
        columns = {col.name: col.type for col in table.columns}

        # Creating Pydantic fields based on SQLite table columns
        fields = {col: (typ, None) for col, typ in columns.items()}
        
        # FAILED to Dynamically creating the Pydantic model using the fields
        # DynamicModel = create_model(table_name, **fields)

        # Defining and returning model inside a function to ensure a fresh namespace
        # FAILED
        # TODO: Figure out how to dynamically create a Pydantic model
        def create_dynamic_model():
            return create_model(table_name, **fields)
    
        return create_dynamic_model()




from abc import ABC, abstractmethod
from typing import List, Union

class BaseEmbedder(ABC):
    @abstractmethod
    def embed(self, text: Union[str, List[str]]) -> Union[List[float], List[List[float]]]:
        pass
    
    @property
    @abstractmethod
    def dimension(self) -> int:
        """
        Return the dimensionality of the embeddings produced by the embedder.
        """
        pass



from pypdf import PdfReader
from .base_loader import BaseLoader

class PdfLoader(BaseLoader):

    def load_data(self, source: str) -> str:
        """
        Load and extract text from a PDF file.

        :param source: Path to the PDF file.
        :return: Extracted text from the PDF.
        """
        with open(source, 'rb') as file:
            reader = PdfReader(file)
            text = ""
            for page in reader.pages:
                text += page.extract_text()
        return text



from .base_loader import BaseLoader

class TextLoader(BaseLoader):

    def load_data(self, source: str) -> str:
        """
        Load plain text data.

        :param source: Plain text data.
        :return: Provided text data (unchanged).
        """
        return source



import re

def detect_data_type(source: str) -> str:
    """
    Detect the type of data based on the provided source.

    :param source: The source (typically a URL) of the data.
    :return: A string indicating the type of data.
    """

    # Check for arXiv papers
    if re.match(r'^https?://arxiv\.org/abs/\d+\.\d+', source):
        return "ARXIV_PAPER"
    
    # Check for PDF files
    if source.endswith('.pdf'):
        return "PDF_FILE"
    
    # Check for SQLite data (using a custom URI scheme)
    if source.startswith('sqlite:///'):
        return "SQLITE_DATA"
    
    raise ValueError(f"Unsupported source: {source}")
